/*
 * Copyright (c) 1989, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Tony Nardo of the Johns Hopkins University/Applied Physics Lab.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char sccsid[] = "@(#)net.c	8.4 (Berkeley) 4/28/95";
#endif /* not lint */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <db.h>
#include <unistd.h>
#include <pwd.h>
#include <utmp.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "finger.h"

#if FASTCTO
#include <signal.h>
#include <setjmp.h>
#endif /* FASTCTO */

#if IPSEC
#include <errno.h>
#endif /* IPSEC */

#if INET6
#include "support.h"
#endif /* INET6 */

#ifdef FASTCTO
static jmp_buf timeout_env;

static void timeout_handler(int i)
{
  longjmp(timeout_env, i);
}
#endif /* FASTCTO */

void
netfinger(name)
	char *name;
{
	extern int lflag;
	register FILE *fp;
	register int c, lastc;
#if !INET6
	struct in_addr defaddr;
#endif /* INET6 */
	struct hostent *hp, def;
	struct servent *sp;
#if INET6
	struct sockaddr_in sin;
#endif /* INET6 */
	int s;
	char *alist[1], *host;
#if INET6
	struct addrinfo req, *ai, *ai2;
	char buf1[46], buf2[16];

	if (!(host = strrchr(name, '@')))
	  return;
	*(host++) = 0;
#if 1
	if (strrchr(name, '@'))
	  return;
#endif /* 1 */

	memset(&req, 0, sizeof(struct addrinfo));
	req.ai_socktype = SOCK_STREAM;
	req.ai_flags |= AI_CANONNAME;

	if (s = getaddrinfo(host, "finger", &req, &ai)) {
          fprintf(stderr, "finger: getaddrinfo: %s: %s\n", host, nrl_gai_strerror(s));
	  return;
	}

#ifdef FASTCTO
	signal(SIGALRM, &timeout_handler);
#endif /* FASTCTO */

	{
	  for (ai2 = ai, s = -1; ai; ai = ai->ai_next) {
	    if ((s = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol)) < 0) {
	      perror("socket");
	      continue;
	    }

#if IPSEC
            if (request)
              if (setsockopt(s, SOL_SOCKET, SO_SECURITY_REQUEST, request, requestlen) < 0) {
                fprintf(stderr, "ftp: setsockopt(..., SO_SECURITY_REQUEST, ...): %s(%d)\n", strerror(errno), errno);
                exit(1);
              };
#endif /* IPSEC */

#if FASTCTO
	    if (setjmp(timeout_env)) {
	      fprintf(stderr, "finger: Connection timed out\n");
	      continue;
	    }

	    alarm(FASTCTO);
#endif /* FASTCTO */

	    if (nrl_getnameinfo(ai->ai_addr, ai->ai_addrlen, buf1, sizeof(buf1), buf2, sizeof(buf2), 0)) {
	      printf("finger: getnameinfo() failed!\n");
	      continue;
	    }

	    printf("Trying %s.%s...\n", buf1, buf2);
	    
	    if (connect(s, ai->ai_addr, ai->ai_addrlen) < 0) {
#ifdef FASTCTO
	      alarm(0);
#endif /* FASTCTO */
	      perror("connect");
	      close(s);
	      s = -1;
	      continue;
	    }
#ifdef FASTCTO
	    alarm(0);
#endif /* FASTCTO */
	    break;
	  }

	  if (s < 0)
	    return;
	}
	printf("[%s]\n", ai->ai_canonname);

	freeaddrinfo(ai2);
#else /* INET6 */
	if (!(host = rindex(name, '@')))
		return;
	*host++ = NULL;
	if (isdigit(*host) && (defaddr.s_addr = inet_addr(host)) != -1) {
		def.h_name = host;
		def.h_addr_list = alist;
		def.h_addr = (char *)&defaddr;
		def.h_length = sizeof(struct in_addr);
		def.h_addrtype = AF_INET;
		def.h_aliases = 0;
		hp = &def;
	} else if (!(hp = gethostbyname(host))) {
		(void)fprintf(stderr,
		    "finger: unknown host: %s\n", host);
		return;
	}
	if (!(sp = getservbyname("finger", "tcp"))) {
		(void)fprintf(stderr, "finger: tcp/finger: unknown service\n");
		return;
	}
	sin.sin_family = hp->h_addrtype;
	bcopy(hp->h_addr, (char *)&sin.sin_addr, hp->h_length);
	sin.sin_port = sp->s_port;
	if ((s = socket(hp->h_addrtype, SOCK_STREAM, 0)) < 0) {
		perror("finger: socket");
		return;
	}

	/* have network connection; identify the host connected with */
	(void)printf("[%s]\n", hp->h_name);
	if (connect(s, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
		perror("finger: connect");
		(void)close(s);
		return;
	}
#endif /* INET6 */

	/* -l flag for remote fingerd  */
	if (lflag)
		write(s, "/W ", 3);
	/* send the name followed by <CR><LF> */
	(void)write(s, name, strlen(name));
	(void)write(s, "\r\n", 2);

	/*
	 * Read from the remote system; once we're connected, we assume some
	 * data.  If none arrives, we hang until the user interrupts.
	 *
	 * If we see a <CR> or a <CR> with the high bit set, treat it as
	 * a newline; if followed by a newline character, only output one
	 * newline.
	 *
	 * Otherwise, all high bits are stripped; if it isn't printable and
	 * it isn't a space, we can simply set the 7th bit.  Every ASCII
	 * character with bit 7 set is printable.
	 */ 
	lastc = 0;
	if ((fp = fdopen(s, "r")) != NULL)
		while ((c = getc(fp)) != EOF) {
			c &= 0x7f;
			if (c == 0x0d) {
				if (lastc == '\r')	/* ^M^M - skip dupes */
					continue;
				c = '\n';
				lastc = '\r';
			} else {
				if (!isprint(c) && !isspace(c))
					c |= 0x40;
				if (lastc != '\r' || c != '\n')
					lastc = c;
				else {
					lastc = '\n';
					continue;
				}
			}
			putchar(c);
		}
	if (lastc != '\n')
		putchar('\n');
	putchar('\n');
	(void)fclose(fp);
}
