/*
 * Copyright 1995 by Randall Atkinson, Bao Phan, and Dan McDonald
 *	All Rights Reserved.  
 *      All Rights under this copyright have been assigned to NRL.
 */

/*----------------------------------------------------------------------
#	@(#)COPYRIGHT	1.1a (NRL) 17 August 1995

COPYRIGHT NOTICE

All of the documentation and software included in this software
distribution from the US Naval Research Laboratory (NRL) are
copyrighted by their respective developers.

This software and documentation were developed at NRL by various
people.  Those developers have each copyrighted the portions that they
developed at NRL and have assigned All Rights for those portions to
NRL.  Outside the USA, NRL also has copyright on the software
developed at NRL. The affected files all contain specific copyright
notices and those notices must be retained in any derived work.

NRL LICENSE

NRL grants permission for redistribution and use in source and binary
forms, with or without modification, of the software and documentation
created at NRL provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:

	This product includes software developed at the Information
	Technology Division, US Naval Research Laboratory.

4. Neither the name of the NRL nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THE SOFTWARE PROVIDED BY NRL IS PROVIDED BY NRL AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL NRL OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation
are those of the authors and should not be interpreted as representing
official policies, either expressed or implied, of the US Naval
Research Laboratory (NRL).

----------------------------------------------------------------------*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <netdb.h>
#include <string.h>
#include <sys/un.h>
#include <netinet/in.h>
#if INET6
#include <netinet6/in6.h>
#endif /* INET6 */

#include "support.h"

#define min(a,b) ((a) <= (b) ? (a) : (b))

static char buffer[6];
static char hextab[] = { '0', '1', '2', '3', '4', '5', '6', '7',
                         '8', '9', 'a', 'b', 'c', 'd', 'e', 'f' };

static struct nrl_nametonum afnametonum[] = {
  { AF_INET, "inet", sizeof(struct sockaddr_in) },
#if INET6
  { AF_INET6, "inet6", sizeof(struct sockaddr_in6) },
#endif /* INET6 */
  { AF_LOCAL, "local", sizeof(struct sockaddr_un) },
  { AF_UNIX, "unix", sizeof(struct sockaddr_un) },
  { 0, NULL, -1 }
};

static struct nrl_nametonum gainametonum[] = {
  { 0, "success", 0 },
  { EAI_BADFLAGS, "invalid value for ai_flags", 0 },
  { EAI_NONAME, "name or service is not known", 0 },
  { EAI_AGAIN, "temporary failure in name resolution", 0 },
  { EAI_FAIL, "non-recoverable failure in name resolution", 0 },
  { EAI_NODATA, "no address associated with name", 0 },
  { EAI_FAMILY, "ai_family not supported", 0 },
  { EAI_SOCKTYPE, "ai_socktype not supported", 0 },
  { EAI_SERVICE, "service not supported for ai_socktype", 0 },
  { EAI_ADDRFAMILY, "address family for name not supported", 0 },
  { EAI_MEMORY, "memory allocation failure", 0 },
  { EAI_SYSTEM, "system error", 0 },
  { 0, NULL, 0 }
};

int nrl_nametonum(tab, name)
struct nrl_nametonum *tab;
char *name;
{
  for (; tab->name; tab++)
    if (!strcasecmp(tab->name, name))
      return(tab->num);

  return -1;
}

char *nrl_numtoname(tab, num)
struct nrl_nametonum *tab;
int num;
{
  for (; tab->name; tab++)
    if (tab->num == num)
      return(tab->name);

  return NULL;
}

int nrl_numtoflags(tab, num)
struct nrl_nametonum *tab;
int num;
{
  for (; tab->name; tab++)
    if (tab->num == num)
      return(tab->flags);

  return -1;
}

int nrl_afnametonum(name)
char *name;
{
  int i = nrl_nametonum(afnametonum, name);
  if ((i < 0) && isdigit(*name))
    i = atoi(name);
  return i;
}

char *nrl_afnumtoname(num)
int num;
{
  char *c = nrl_numtoname(afnametonum, num);
 
  if (!c) {
    buffer[0] = 'a';
    buffer[1] = 'f';
    buffer[2] = ' ';
    buffer[3] = hextab[(num & 0xf0) >> 4];
    buffer[4] = hextab[num & 0x0f];
    buffer[5] = 0;
  }

  return c;
}

char *nrl_gai_strerror(num)
int num;
{
  char *c = nrl_numtoname(gainametonum, num);
 
  if (!c) {
    buffer[0] = '?';
    buffer[1] = ' ';
    buffer[2] = hextab[(num & 0xf0) >> 4];
    buffer[3] = hextab[num & 0x0f];
    buffer[4] = 0;
  }

  return c;
}

#if INET6
const struct in6_addr in6addr_any = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 };
const struct in6_addr in6addr_loopback = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1 };

int inet6_isipv4mapped(const struct in6_addr *addr)
{
  return IN6_IS_ADDR_V4MAPPED(addr);
}
#endif /* INET6 */

static char *domain = NULL;
static char domainbuffer[MAXHOSTNAMELEN] = "";

char *nrl_domainname(void)
{
  static int first = 1;

  if (first) {
    char *c, buf[MAXHOSTNAMELEN];
    struct hostent *h;

    first = 0;

    if ((h = gethostbyname("localhost")) && (c = strchr(h->h_name, '.')))
      return strcpy(domain = domainbuffer, ++c);

    if (!gethostname(domainbuffer, sizeof(domainbuffer))) {
      if (c = strchr(domainbuffer, '.'))
        return (domain = ++c);

      if ((h = gethostbyname(domainbuffer)) && (c = strchr(h->h_name, '.')))
        return strcpy(domain = domainbuffer, ++c);
    }

    {
      struct in_addr in_addr;

      in_addr.s_addr = htonl(0x7f000001);

      if ((h = gethostbyaddr((const char *)&in_addr, sizeof(struct in_addr), AF_INET)) && (c = strchr(h->h_name, '.')))
        return strcpy(domain = domainbuffer, ++c);
    }

    return NULL;
  };

  return domain;
};

int nrl_build_samask(struct sockaddr *mask, int af, int len)
{
  u_int8_t *p;
  if (len < 0)
    return -1;
  switch(af) {
#if INET6
    case AF_INET6:
      if (len > 128)
      return -1;
      p = (u_int8_t *)&(((struct sockaddr_in6 *)mask)->sin6_addr);
      memset(mask, 0, sizeof(struct sockaddr_in6));
#if 0
      ((struct sockaddr_in6 *)mask)->sin6_port = ~1;
#endif /* 0 */
      break;
#endif /* INET6 */
    case AF_INET:
      if (len > 32)
      return -1;
      p = (u_int8_t *)&(((struct sockaddr_in *)mask)->sin_addr);
      memset(mask, 0, sizeof(struct sockaddr_in));
#if 0
      ((struct sockaddr_in *)mask)->sin_port = ~1;
#endif /* 0 */
      break;
    default:
      return -1;
  };
#if SALEN
  mask->sa_len = ~1;
#endif /* SALEN */
  mask->sa_family = ~1;
  while(len >= 8) {
    *(p++) = 0xff;
    len -= 8;
  };
  while(len--)
    *p = (*p << 1) | 1;
  return 0;
};
int nrl_prefix_length(struct sockaddr *mask, int af)
{
  int len = 0;
  int i, l;
  u_int8_t *p;

  switch(af) {
  case AF_INET:
#if SALEN
    l = min(mask->sa_len - 4, sizeof(struct in_addr));
#else /* SALEN */
    l = sizeof(struct in_addr);
#endif /* SALEN */
    p = (u_int8_t *)&((struct sockaddr_in *)mask)->sin_addr;
    break;
#if INET6
  case AF_INET6:
#if SALEN
    l = min(mask->sa_len - 8, sizeof(struct in6_addr));
#else /* SALEN */
    l = sizeof(struct in6_addr);
#endif /* SALEN */
    p = (u_int8_t *)&((struct sockaddr_in6 *)mask)->sin6_addr;
    break;
#endif /* INET6 */
  default:
    return -1;
  }

  while ((l > 0) && (*p == 0xff)) {
    len += 8;
    p++;
    l--;
  }
  if (l < 1)
    return len;

  if (*p) {
    i = *(p++);
    l--;
    while(i & 0x80) {
      len++;
      i = (i << 1) & 0xff;
    }
    if (i)
      return -1;
  }

  while(l-- > 0)
    if (*(p++))
      return -1;

  return len;
}

#if !SALEN
int nrl_sa_len(struct sockaddr *sa)
{
  int i;

  if ((i = nrl_numtoflags(afnametonum, sa->sa_family)) < 0)
    return 0;
  return i;
}
#endif /* !SALEN */
